## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg as spla
from ridge import ridge_base
from ols import ols_base
from kernel_class import *

import sys
if sys.version >= '3':
    from . import kernel
else:
    import kernel

__all__ = ['LDA', 'SRDA', 'KFDA', 'PCA', 'PCAFast', 'KPCA']


def proj(u, v):
    """(<v, u> / <u, u>) u
    """

    return (np.dot(v, u) / np.dot(u, u)) * u


def gso(v, norm=False):
    """Gram-Schmidt orthogonalization.
    Vectors v_1, ..., v_k are stored by rows.
    """
    
    for j in range(v.shape[0]):
        for i in range(j):
            v[j] = v[j] - proj(v[i], v[j])
        
        if norm:
            v[j] /= np.linalg.norm(v[j])


def lda(xarr, yarr):
    """Linear Discriminant Analysis.
    
    Returns the transformation matrix `coeff` (P, C-1),
    where `x` is a matrix (N,P) and C is the number of
    classes. Each column of `x` represents a variable, 
    while the rows contain observations. Each column of 
    `coeff` contains coefficients for one transformation
    vector.
    
    Sample(s) can be embedded into the C-1 dimensional space
    by z = x coeff (z = np.dot(x, coeff)).

    :Parameters:
       x : 2d array_like object (N, P)
          data matrix
       y : 1d array_like object integer (N)
          class labels
    
    :Returns:
       coeff: 2d numpy array (P, P)
          transformation matrix.
    """

    n, p = xarr.shape[0], xarr.shape[1]
    labels = np.unique(yarr)
    
    sw = np.zeros((p, p), dtype=np.float)   
    for i in labels:
        idx = np.where(yarr==i)[0]
        sw += np.cov(xarr[idx], rowvar=0) * \
            (idx.shape[0] - 1)
    st = np.cov(xarr, rowvar=0) * (n - 1)

    sb = st - sw
    evals, evecs = spla.eig(sb, sw, overwrite_a=True,
                            overwrite_b=True)
    idx = np.argsort(evals)[::-1]
    evecs = evecs[:, idx]
    evecs = evecs[:, :labels.shape[0]-1]
    
    return evecs


def srda(xarr, yarr, alpha):
    """Spectral Regression Discriminant Analysis.

    Returns the (P, C-1) transformation matrix, where 
    `x` is a matrix (N,P) and C is the number of classes.
    Each column of `x` represents a variable, while the 
    rows contain observations. `x` must be centered 
    (subtracting the empirical mean vector from each column 
    of`x`).

    Sample(s) can be embedded into the C-1 dimensional space
    by z = x coeff (z = np.dot(x, coeff)).

    :Parameters:
       x : 2d array_like object
          training data (N, P)
       y : 1d array_like object integer
          target values (N)
       alpha : float (>=0)
          regularization parameter

    :Returns:
       coeff : 2d numpy array (P, C-1)
          tranformation matrix
    """

    # Point 1 in section 4.2
    yu = np.unique(yarr)
    yk = np.zeros((yu.shape[0]+1, yarr.shape[0]), dtype=np.float)
    yk[0] = 1.
    for i in range(1, yk.shape[0]):
        yk[i][yarr==yu[i-1]] = 1.
    gso(yk, norm=False) # orthogonalize yk
    yk = yk[1:-1]
    
    # Point 2 in section 4.2
    ak = np.empty((yk.shape[0], xarr.shape[1]), dtype=np.float)
    for i in range(yk.shape[0]):
        ak[i] = ridge_base(xarr, yk[i], alpha)

    return ak.T


def pca(xarr, method='svd'):
    """Principal Component Analysis.
    
    Returns the principal component coefficients `coeff`(K,K) 
    and the corresponding eigenvalues (K) of the covariance 
    matrix of `x` (N,P) sorted by decreasing eigenvalue, where 
    K=min(N,P). Each column of `x` represents a variable,  
    while the rows contain observations. Each column of `coeff` 
    contains coefficients for one principal component.
    
    Sample(s) can be embedded into the M (<=K) dimensional
    space by z = x coeff_M (z = np.dot(x, coeff[:, :M])).

    :Parameters:
       x : 2d numpy array (N, P)
          data matrix
       method : str
          'svd' or 'cov'
    
    :Returns:
       coeff, evals : 2d numpy array (K, K), 1d numpy array (K)
          principal component coefficients (eigenvectors of
          the covariance matrix of x) and eigenvalues sorted by 
          decreasing eigenvalue.
    """


    n, p = xarr.shape
    
    if method == 'svd':
        x_h = (xarr - np.mean(xarr, axis=0)) / np.sqrt(n - 1)
        u, s, v = np.linalg.svd(x_h.T, full_matrices=False)
        evecs = u
        evals = s**2
    elif method == 'cov':
        k = np.min((n, p))
        C = np.cov(xarr, rowvar=0)
        evals, evecs = np.linalg.eigh(C)
        idx = np.argsort(evals)[::-1]
        evecs = evecs[:, idx]
        evals = evals[idx]
        evecs = evecs[:, :k]
        evals = evals[:k]
    else:
        raise ValueError("method must be 'svd' or 'cov'")

    return evecs, evals
        

def pca_fast(xarr, m, eps):
    """Fast principal component analysis using the fixed-point
    algorithm.
    
    Returns the first `m` principal component coefficients
    `coeff` (P, M). Each column of `x` represents a variable,  
    while the rows contain observations. Each column of `coeff` 
    contains coefficients for one principal component.

    Sample(s) can be embedded into the m (<=P) dimensional space 
    by z = x coeff (z = np.dot(X,  coeff)).

    :Parameters:
       x : 2d numpy array (N, P)
          data matrix
       m : integer (0 < m <= P) 
          the number of principal axes or eigenvectors required
       eps : float (> 0)
          tolerance error
    
    :Returns:
       coeff : 2d numpy array (P, H)
          principal component coefficients
    """
    
    m = int(m)

    np.random.seed(0)
    evecs = np.random.rand(m, xarr.shape[1])

    C = np.cov(xarr, rowvar=0)    
    for i in range(0, m):
        while True:
            evecs_old = np.copy(evecs[i])
            evecs[i] = np.dot(C, evecs[i])
            
            # Gram-Schmidt orthogonalization
            a = np.dot(evecs[i], evecs[:i].T).reshape(-1, 1)
            b = a  * evecs[:i]
            evecs[i] -= np.sum(b, axis=0) # if i=0 sum is 0
            
            # Normalization
            evecs[i] = evecs[i] / np.linalg.norm(evecs[i])
            
            # convergence criteria
            if np.abs(np.dot(evecs[i], evecs_old) - 1) < eps:
                break

    return evecs.T
      

def lda_fast(xarr, yarr):
    """Fast implementation of Linear Discriminant Analysis.
    
    Returns the (P, C-1) transformation matrix, where 
    `x` is a centered matrix (N,P) and C is the number of classes.
    Each column of `x` represents a variable, while the 
    rows contain observations. `x` must be centered 
    (subtracting the empirical mean vector from each column 
    of`x`).

    :Parameters:
       x : 2d array_like object
          training data (N, P)
       y : 1d array_like object integer
          target values (N)
    
    :Returns:
       A : 2d numpy array (P, C-1)
          tranformation matrix
    """

    yu = np.unique(yarr)
    yk = np.zeros((yu.shape[0]+1, yarr.shape[0]), dtype=np.float)
    yk[0] = 1.
    for i in range(1, yk.shape[0]):
        yk[i][yarr==yu[i-1]] = 1.
    gso(yk, norm=False) # orthogonalize yk
    yk = yk[1:-1]
    
    ak = np.empty((yk.shape[0], xarr.shape[1]), dtype=np.float)
    for i in range(yk.shape[0]):
        ak[i], _ = ols_base(xarr, yk[i], -1)

    return ak.T


def kpca(K):
    """Kernel Principal Component Analysis, PCA in 
    a kernel-defined feature space making use of the
    dual representation.
    
    Returns the kernel principal component coefficients 
    `coeff` (N, N) computed as :math:`\lambda^{-1/2} \mathbf{v}_j`
    where :math:`\lambda` and :math:`\mathbf{v}` are the ordered
    eigenvalues and the corresponding eigenvector of the centered 
    kernel matrix K.
    
    Sample(s) can be embedded into the G (<=N) dimensional space
    by z = K coeff_G (z = np.dot(K, coeff[:, :G])).

    :Parameters:
       K: 2d array_like object (N,N)
          precomputed centered kernel matrix
        
    :Returns:
       coeff, evals: 2d numpy array (N,N), 1d numpy array (N)
          kernel principal component coefficients, eigenvalues
          sorted by decreasing eigenvalue.
    """
    
    evals, evecs = np.linalg.eigh(K)
    idx = np.argsort(evals)
    idx = idx[::-1]
    evecs = evecs[:, idx]
    evals = evals[idx]
    
    for i in range(len(evals)):
        evecs[:, i] /= np.sqrt(evals[i])
   
    return evecs, evals


def kfda(Karr, yarr, lmb=0.001):
    """Kernel Fisher Discriminant Analysis.
    
    Returns the transformation matrix `coeff` (N,1),
    where `K` is a the kernel matrix (N,N) and y
    is the class labels (the alghoritm works only with 2
    classes).
   
    :Parameters:
       K: 2d array_like object (N, N)
          precomputed kernel matrix
       y : 1d array_like object integer (N)
          class labels
       lmb : float (>= 0.0)
          regularization parameter

    :Returns:
       coeff: 2d numpy array (N,1)
          kernel fisher coefficients.
    """  

    labels = np.unique(yarr)
    n = yarr.shape[0]

    idx1 = np.where(yarr==labels[0])[0]
    idx2 = np.where(yarr==labels[1])[0]
    n1 = idx1.shape[0]
    n2 = idx2.shape[0]
    
    K1, K2 = Karr[:, idx1], Karr[:, idx2]
    
    N1 = np.dot(np.dot(K1, np.eye(n1) - (1 / float(n1))), K1.T)
    N2 = np.dot(np.dot(K2, np.eye(n2) - (1 / float(n2))), K2.T)
    N = N1 + N2 + np.diag(np.repeat(lmb, n))

    M1 = np.sum(K1, axis=1) / float(n1)
    M2 = np.sum(K2, axis=1) / float(n2)
    M = M1 - M2
    
    coeff = np.linalg.solve(N, M).reshape(-1, 1)
            
    return coeff
 

class LDA:
    """Linear Discriminant Analysis.
    """
    
    def __init__(self, method='cov'):
        """Initialization.
        
        :Parameters:
           method : str
              'cov' or 'fast'
        """
        
        self._coeff = None
        self._mean = None

        if method not in ['cov', 'fast']:
            raise ValueError("method must be 'cov' or 'fast'")

        self._method = method

    def learn(self, x, y):
        """Computes the transformation matrix.
        `x` is a matrix (N,P) and `y` is a vector containing
        the class labels. Each column of `x` represents a 
        variable, while the rows contain observations.
        """

        xarr = np.asarray(x, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)
        
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")
        
        if xarr.shape[0] != yarr.shape[0]:
            raise ValueError("x, y shape mismatch")
        
        self._mean = np.mean(xarr, axis=0)
        
        if self._method == 'cov':
            self._coeff = lda(xarr, yarr)
        elif self._method == 'fast':
            self._coeff = lda_fast(xarr-self._mean, yarr)

    def transform(self, t):
        """Embed `t` (M,P) into the C-1 dimensional space.
        Returns a (M,C-1) matrix.
        """
        if self._coeff is None:
            raise ValueError("no model computed")

        tarr = np.asarray(t, dtype=np.float)
        
        try:
            return np.dot(tarr-self._mean, self._coeff)
        except:
            ValueError("t, coeff: shape mismatch")

    def coeff(self):
        """Returns the tranformation matrix (P,C-1), where
        C is the number of classes. Each column contains 
        coefficients for one transformation vector.
        """

        return self._coeff


class SRDA:
    """Spectral Regression Discriminant Analysis.
    """
    
    def __init__(self, alpha=0.001):
        """Initialization.

        :Parameters:
           alpha : float (>=0)
              regularization parameter
        """
        
        self._coeff = None
        self._mean = None
        self._alpha = alpha

    def learn(self, x, y):
        """Computes the transformation matrix.
        `x` is a matrix (N,P) and `y` is a vector containing
        the class labels. Each column of `x` represents a 
        variable, while the rows contain observations.
        """

        xarr = np.asarray(x, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)
        
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")
        
        if xarr.shape[0] != yarr.shape[0]:
            raise ValueError("x, y shape mismatch")
        
        self._mean = np.mean(xarr, axis=0)
        self._coeff = srda(xarr-self._mean, yarr, self._alpha)

    def transform(self, t):
        """Embed t (M,P) into the C-1 dimensional space.
        Returns a (M,C-1) matrix.
        """

        if self._coeff is None:
            raise ValueError("no model computed")

        tarr = np.asarray(t, dtype=np.float)
        
        try:
            return np.dot(tarr-self._mean, self._coeff)
        except:
            ValueError("t, coeff: shape mismatch")

    def coeff(self):
        """Returns the tranformation matrix (P,C-1), where
        C is the number of classes. Each column contains 
        coefficients for one transformation vector.
        """

        return self._coeff


class KFDA:
    """Kernel Fisher Discriminant Analysis.
    """
    
    def __init__(self, lmb=0.001, kernel=None):
        """Initialization.
        
        :Parameters:
           lmb : float (>= 0.0)
              regularization parameter
           kernel : None or mlpy.Kernel object.
              if kernel is None, K and Kt in .learn()
              and in .transform() methods must be precomputed kernel 
              matricies, else K and Kt must be training (resp. 
              test) data in input space.
        """

        
        if kernel is not None:
            if not isinstance(kernel, Kernel):
                raise ValueError("kernel must be None or a mlpy.Kernel object")
    
        self._kernel = kernel
        self._x = None
        self._coeff = None
        self._lmb = lmb

    def learn(self, K, y):
        """Computes the transformation vector.

        :Parameters:
           K: 2d array_like object
              precomputed training kernel matrix (if kernel=None);
              training data in input space (if kernel is a Kernel object)
           y : 1d array_like object integer (N)
              class labels (only two classes)
        """

        Karr = np.array(K, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)
        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")

        if self._kernel is None:
            if Karr.shape[0] != Karr.shape[1]:
                raise ValueError("K must be a square matrix")
        else:
            self._x = Karr.copy()
            Karr = self._kernel.kernel(Karr, Karr)

        labels = np.unique(yarr)
        if labels.shape[0] != 2:
            raise ValueError("number of classes must be = 2")
        
        self._coeff = kfda(Karr, yarr, self._lmb)

    def transform(self, Kt):
        """Embed Kt into the 1d kernel fisher space.
        
        :Parameters:
           Kt : 1d or 2d array_like object
              precomputed test kernel matrix. (if kernel=None);
              test data in input space (if kernel is a Kernel object).
        """

        if self._coeff is None:
            raise ValueError("no model computed")

        Ktarr = np.asarray(Kt, dtype=np.float)
        if self._kernel is not None:
            Ktarr = self._kernel.kernel(Ktarr, self._x)

        try:
            return np.dot(Ktarr, self._coeff)
        except:
            ValueError("Kt, coeff: shape mismatch")

    def coeff(self):
        """Returns the tranformation vector (N,1).
        """

        return self._coeff


class PCA:
    """Principal Component Analysis.
    """
    
    def __init__(self, method='svd', whiten=False):
        """Initialization.
        
        :Parameters:
           method : str
              method, 'svd' or 'cov'
           whiten : bool
              whitening. The eigenvectors will be scaled
              by eigenvalues**-(1/2)
        """
        
        self._coeff = None
        self._coeff_inv = None
        self._evals = None
        self._mean = None
        self._method = method
        self._whiten = whiten        

    def learn(self, x):
        """Compute the principal component coefficients.
        `x` is a matrix (N,P). Each column of `x` represents a 
        variable, while the rows contain observations.
        """

        xarr = np.asarray(x, dtype=np.float)
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        self._mean = np.mean(xarr, axis=0)
        self._coeff, self._evals = pca(x, method=self._method)

        if self._whiten:
            self._coeff_inv = np.empty((self._coeff.shape[1], 
                self._coeff.shape[0]), dtype=np.float)
            
            for i in range(len(self._evals)):
                eval_sqrt = np.sqrt(self._evals[i])
                self._coeff_inv[i] = self._coeff[:, i] * \
                    eval_sqrt
                self._coeff[:, i] /= eval_sqrt
        else:
            self._coeff_inv = self._coeff.T

    def transform(self, t, k=None):
        """Embed `t` (M,P) into the k dimensional subspace.
        Returns a (M,K) matrix. If `k` =None will be set to 
        min(N,P)
        """

        if self._coeff is None:
            raise ValueError("no PCA computed")
        
        if k == None:
            k = self._coeff.shape[1]

        if k < 1 or k > self._coeff.shape[1]:
            raise ValueError("k must be in [1, %d] or None" % \
                                 self._coeff.shape[1])

        tarr = np.asarray(t, dtype=np.float)

        try:
            return np.dot(tarr-self._mean, self._coeff[:, :k])
        except:
            raise ValueError("t, coeff: shape mismatch")
            
    def transform_inv(self, z):
        """Transform data back to its original space,
        where `z` is a (M,K) matrix. Returns a (M,P) matrix.
        """

        if self._coeff is None:
            raise ValueError("no PCA computed")

        zarr = np.asarray(z, dtype=np.float)

        return np.dot(zarr, self._coeff_inv[:zarr.shape[1]]) +\
            self._mean
        
    def coeff(self):
        """Returns the tranformation matrix (P,L), where
        L=min(N,P), sorted by decreasing eigenvalue.
        Each column contains coefficients for one principal 
        component.
        """
        
        return self._coeff
    
    def coeff_inv(self):
        """Returns the inverse of tranformation matrix (L,P),
        where L=min(N,P), sorted by decreasing eigenvalue.
        """
        
        return self._coeff_inv

    def evals(self):
        """Returns sorted eigenvalues (L), where L=min(N,P).
        """
        
        return self._evals


class PCAFast:
    """Fast Principal Component Analysis.
    """
    
    def __init__(self, k=2, eps=0.01):
        """Initialization.
        
        :Parameters:
           k : integer
              the number of principal axes or eigenvectors required
           eps : float (> 0)
              tolerance error
        """
        
        self._coeff = None
        self._coeff_inv = None
        self._mean = None
        self._k = k
        self._eps = eps        

    def learn(self, x):
        """Compute the firsts `k` principal component coefficients.
        `x` is a matrix (N,P). Each column of `x` represents a 
        variable, while the rows contain observations.
        """

        xarr = np.asarray(x, dtype=np.float)
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        self._mean = np.mean(xarr, axis=0)
        self._coeff = pca_fast(xarr, m=self._k, eps=self._eps)
        self._coeff_inv = self._coeff.T

    def transform(self, t):
        """Embed t (M,P) into the `k` dimensional subspace.
        Returns a (M,K) matrix.
        """

        if self._coeff is None:
            raise ValueError("no PCA computed")

        tarr = np.asarray(t, dtype=np.float)

        try:
            return np.dot(tarr-self._mean, self._coeff)
        except:
            raise ValueError("t, coeff: shape mismatch")
            
    def transform_inv(self, z):
        """Transform data back to its original space,
        where `z` is a (M,K) matrix. Returns a (M,P) matrix.
        """

        if self._coeff is None:
            raise ValueError("no PCA computed")

        zarr = np.asarray(z, dtype=np.float)
        return np.dot(zarr, self._coeff_inv) + self._mean
        
    def coeff(self):
        """Returns the tranformation matrix (P,K) sorted by 
        decreasing eigenvalue.
        Each column contains coefficients for one principal 
        component.
        """
        
        return self._coeff
    
    def coeff_inv(self):
        """Returns the inverse of tranformation matrix (K,P),
        sorted by decreasing eigenvalue.
        """
        
        return self._coeff_inv


class KPCA:
    """Kernel Principal Component Analysis.
    """
    
    def __init__(self, kernel=None):
        """Initialization.
        
        :Parameters:
           kernel : None or mlpy.Kernel object.
              if kernel is None, K and Kt in .learn()
              and in .transform() methods must be precomputed kernel 
              matricies, else K and Kt must be training (resp. 
              test) data in input space.
        """

        if kernel is not None:
            if not isinstance(kernel, Kernel):
                raise ValueError("kernel must be None or a mlpy.Kernel object")
        
        self._coeff = None
        self._evals = None
        self._K = None
        self._kernel = kernel
        self._x = None

    def learn(self, K):
        """Compute the kernel principal component coefficients.

        :Parameters:
           K: 2d array_like object
              precomputed training kernel matrix (if kernel=None);
              training data in input space (if kernel is a Kernel object)
        """

        Karr = np.asarray(K, dtype=np.float)
        if Karr.ndim != 2:
            raise ValueError("K must be a 2d array_like object")
        
        if self._kernel is None:
            if Karr.shape[0] != Karr.shape[1]:
                raise ValueError("K must be a square matrix")
        else:
            self._x = Karr.copy()
            Karr = self._kernel.kernel(Karr, Karr)

        self._K = Karr.copy()
        Karr = kernel.kernel_center(Karr, Karr)
        self._coeff, self._evals = kpca(Karr)
       
    def transform(self, Kt, k=None):
        """Embed Kt into the `k` dimensional subspace.
        
        :Parameters:
           Kt : 1d or 2d array_like object
              precomputed test kernel matrix. (if kernel=None);
              test data in input space (if kernel is a Kernel object).
        """

        if self._coeff is None:
            raise ValueError("no KPCA computed")
        
        if k == None:
            k = self._coeff.shape[1]

        if k < 1 or k > self._coeff.shape[1]:
            raise ValueError("k must be in [1, %d] or None" % \
                                 self._coeff.shape[1])

        Ktarr = np.asarray(Kt, dtype=np.float)
        if self._kernel is not None:
            Ktarr = self._kernel.kernel(Ktarr, self._x)
        
        Ktarr = kernel.kernel_center(Ktarr, self._K)

        try:
            return np.dot(Ktarr, self._coeff[:, :k])
        except:
            raise ValueError("Kt, coeff: shape mismatch")
 
    def coeff(self):
        """Returns the tranformation matrix (N,N) sorted by 
        decreasing eigenvalue.
        """
        
        return self._coeff
    
    def evals(self):
        """Returns sorted eigenvalues (N).
        """
        
        return self._evals
   



cdef extern from "cdtw.h":
    
    ctypedef struct Path:
       int k
       int *px
       int *py
       
    double std(double *x, double *y, int n, int m, double *cost, int squared)
    int path(double *cost, int n, int m, int startx, int starty, Path *p)
    void subsequence(double *x, double *y, int n, int m, double *cost)
    int subsequence_path(double *cost, int n, int m, int starty, Path *p)
    



## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['rfe_kfda', 'rfe_w2']

import numpy as np
from kernel_class import *
from da import KFDAC


# used in rfe_kfda
def rayleigh(x, kernel, lmb, alpha, idx1, idx2):
    R = np.empty(x.shape[1], dtype=np.float)
    idx = np.ones(x.shape[1], dtype=bool)
    for i in range(x.shape[1]):
        idx[i] = False
        xi = x[:, idx]
        n = xi.shape[0]
        n1, n2 = idx1.shape[0], idx2.shape[0]
        K = kernel.kernel(xi, xi)
        K1, K2 = K[:, idx1], K[:, idx2]
        N1 = np.dot(np.dot(K1, np.eye(n1) - \
                 (1 / float(n1))), K1.T)
        N2 = np.dot(np.dot(K2, np.eye(n2) - \
                 (1 / float(n2))), K2.T)
        N = N1 + N2 + np.diag(np.repeat(lmb, n))
        m1 = np.sum(K1, axis=1) / float(n1)
        m2 = np.sum(K2, axis=1) / float(n2)
        d = m1 - m2
        M = np.dot(d.reshape(-1, 1), d.reshape(1, -1))
        R[i] = np.dot(np.dot(alpha, M), alpha.reshape(-1, 1)) / \
            np.dot(np.dot(alpha, N), alpha.reshape(-1, 1))
        idx[i] = True
        
    return R


def rfe_kfda(x, y, p, lmb, kernel):
    """KFDA-RFE algorithm based on the Rayleigh coefficient
    proposed in [Louw06]_. The algorithm works with only two
    classes.
    
    .. [Louw06] N Louw and S J Steel. Variable selection in kernel Fisher discriminant analysis by means of recursive feature elimination. Journal Computational Statistics & Data Analysis, 2006.
    
    :Parameters:
       x: 2d array_like object (N,P)
          training data
       y : 1d array_like object integer (N)
          class labels (only two classes)
       p : float [0.0, 1.0]
          percentage  of features (upper rounded) to remove
          at each iteration (p=0 one variable)
       lmb : float (>= 0.0)
          regularization parameter
       kernel : mlpy.Kernel object.
          kernel.

    :Returns:
       ranking : 1d numpy array int
          feature ranking. ranking[i] contains the feature index ranked
          in i-th position.
    """

    if (p < 0.0) or (p > 1.0):
        raise ValueError("parameter p must be in [0.0, 1.0]")

    xarr = np.asarray(x, dtype=np.float)
    yarr = np.asarray(y, dtype=np.int)

    if xarr.ndim != 2:
        raise ValueError("x must be a 2d array_like object")

    if yarr.ndim != 1:
        raise ValueError("y must be an 1d array_like object")

    if xarr.shape[0] != yarr.shape[0]:
        raise ValueError("x, y shape mismatch")

    if not isinstance(kernel, Kernel):
        raise ValueError("kernel must be None or a mlpy.Kernel object")
    
    labels = np.unique(yarr)
    if labels.shape[0] != 2:
        raise ValueError("number of classes must be = 2")
    
    idx1 = np.where(yarr==labels[0])[0]
    idx2 = np.where(yarr==labels[1])[0]

    kfda = KFDAC(lmb=lmb, kernel=kernel)
    idxglobal = np.arange(xarr.shape[1], dtype=np.int)
    ranking = []

    while True:
        nelim = np.max((int(np.ceil(idxglobal.shape[0] * p)), 1))
        xi = xarr[:, idxglobal]
        # compute alpha
        kfda.learn(xi, yarr)
        alpha = kfda.alpha()
        # compute the rayleigh coeffs
        R = rayleigh(xi, kernel, lmb, alpha, idx1, idx2)
        # sorted indexes (descending order)
        idxsorted = np.argsort(R)[::-1]
        # indexes to remove
        idxelim = idxglobal[idxsorted[:nelim]][::-1]
        ranking.insert(0, idxelim)
        # update idxglobal
        idxglobal = idxglobal[idxsorted[nelim:]]
        idxglobal.sort()
        
        if len(idxglobal) <= 1:
            ranking.insert(0, idxglobal)
            break
    
    return np.concatenate(ranking)


def rfe_w2(x, y, p, classifier):
    """RFE algorithm, where the ranking criteria is w^2,
    described in [Guyon02]_. `classifier` must be an linear classifier 
    with learn() and w() methods.
        
    .. [Guyon02] I Guyon, J Weston, S Barnhill and V Vapnik. Gene Selection for Cancer Classification using Support Vector Machines. Machine Learning, 2002.
    
    :Parameters:
       x: 2d array_like object (N,P)
          training data
       y : 1d array_like object integer (N)
          class labels (only two classes)
       p : float [0.0, 1.0]
          percentage  of features (upper rounded) to remove
          at each iteration (p=0 one variable)
       classifier : object with learn() and w() methods
          object

    :Returns:
       ranking : 1d numpy array int
          feature ranking. ranking[i] contains the feature index ranked
          in i-th position.
    """

    
    if (p < 0.0) or (p > 1.0):
        raise ValueError("parameter p must be in [0.0, 1.0]")

    if not (hasattr(classifier, 'learn') and hasattr(classifier, 'w')):
        raise ValueError("parameter classifier must have learn() and w() methods")

    xarr = np.asarray(x, dtype=np.float)
    yarr = np.asarray(y, dtype=np.int)

    if xarr.ndim != 2:
        raise ValueError("x must be a 2d array_like object")

    if yarr.ndim != 1:
        raise ValueError("y must be an 1d array_like object")

    if xarr.shape[0] != yarr.shape[0]:
        raise ValueError("x, y shape mismatch")
    
    labels = np.unique(yarr)
    if labels.shape[0] != 2:
        raise ValueError("number of classes must be = 2")
    
    idxglobal = np.arange(xarr.shape[1], dtype=np.int)
    ranking = []

    while True:
        nelim = np.max((int(np.ceil(idxglobal.shape[0] * p)), 1))
        xi = xarr[:, idxglobal]
        classifier.learn(xi, yarr)
        w = classifier.w()
        idxsorted = np.argsort(w**2)
        # indexes to remove
        idxelim = idxglobal[idxsorted[:nelim]][::-1]
        ranking.insert(0, idxelim)
        # update idxglobal
        idxglobal = idxglobal[idxsorted[nelim:]]
        idxglobal.sort()
        
        if len(idxglobal) <= 1:
            ranking.insert(0, idxglobal)
            break
    
    return np.concatenate(ranking)



## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


__all__ = ["Parzen"]

import numpy as np


class Parzen:
    """Parzen based classifier (binary).
    """

    def __init__(self, kernel=None):
        """Initialization.

        :Parameters:
           kernel : None or mlpy.Kernel object.
              if kernel is None, K and Kt in .learn()
              and in .pred() methods must be precomputed kernel 
              matricies, else K and Kt must be training (resp. 
              test) data in input space.
        """

        self._alpha = None
        self._b = None
        self._labels = None
        self._kernel = kernel
        self._x = None
                
    def learn(self, K, y):
        """Compute alpha and b.

        Parameters:
           K: 2d array_like object
              precomputed training kernel matrix (if kernel=None);
              training data in input space (if kernel is a Kernel object)
           y : 1d array_like object (N)
              target values
        """

        K_arr = np.asarray(K, dtype=np.float)
        y_arr = np.asarray(y, dtype=np.int)

        if K_arr.ndim != 2:
            raise ValueError("K must be a 2d array_like object")

        if y_arr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")
        
        if K_arr.shape[0] != y_arr.shape[0]:
            raise ValueError("K, y shape mismatch")

        if self._kernel is None:
            if K_arr.shape[0] != K_arr.shape[1]:
                raise ValueError("K must be a square matrix")
        else:
            self._x = K_arr.copy()
            K_arr = self._kernel.kernel(K_arr, K_arr)
        
        self._labels = np.unique(y_arr)
        if self._labels.shape[0] != 2:
            raise ValueError("number of classes != 2")

        ynew = np.where(y_arr==self._labels[0], -1., 1.)
        n = K_arr.shape[0]
        
        # from Kernel Methods for Pattern Analysis
        # Algorithm 5.6
        
        nplus = np.sum(ynew==1)
        nminus = n - nplus
        alphaplus = np.where(ynew==1, nplus**-1, 0)
        alphaminus = np.where(ynew==-1, nminus**-1, 0)
        self._b = -0.5 * (np.dot(np.dot(alphaplus, K_arr), alphaplus) - \
                         np.dot(np.dot(alphaminus, K_arr), alphaminus))
        self._alpha = alphaplus - alphaminus

    def pred(self, Kt):
        """Compute the predicted class.

        :Parameters:
           Kt : 1d or 2d array_like object
              precomputed test kernel matrix. (if kernel=None);
              test data in input space (if kernel is a Kernel object).

        :Returns:
           p : integer or 1d numpy array
              predicted class
        """

        if self._alpha is None:
            raise ValueError("no model computed; run learn()")

        Kt_arr = np.asarray(Kt, dtype=np.float)
        if self._kernel is not None:
            Kt_arr = self._kernel.kernel(Kt_arr, self._x)

        try:
            s = np.sign(np.dot(self._alpha, Kt_arr.T) + self._b)
        except ValueError:
            raise ValueError("Kt, alpha: shape mismatch")

        return np.where(s==-1, self._labels[0], self._labels[1]) \
            .astype(np.int)
          
    def alpha(self):
        """Return alpha.
        """
        return self._alpha

    def b(self):
        """Return b.
        """
        return self._b
    
    def labels(self):
        """Outputs the name of labels.
        """
        
        return self._labels
                        



## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["Kernel", "KernelLinear", "KernelPolynomial", "KernelGaussian",
           "KernelExponential", "KernelSigmoid"]


import sys
if sys.version >= '3':
    from . import kernel
else:
    import kernel

class Kernel:
    """Base class for kernels.
    """

    pass

class KernelLinear(Kernel):
    """Linear kernel, t_i' x_j.
    """
    def __init__(self):
        pass
    def kernel(self, t, x):
        return kernel.kernel_linear(t, x)

class KernelPolynomial(Kernel):
    """Polynomial kernel, (gamma t_i' x_j + b)^d.
    """
    def __init__(self, gamma=1.0, b=1.0, d=2.0):
        self.gamma = gamma
        self.b = b
        self.d = d
    def kernel(self, t, x):
        return kernel.kernel_polynomial(t, x,
                   self.gamma, self.b, self.d)

class KernelGaussian(Kernel):
    """Gaussian kernel, exp(-||t_i - x_j||^2 / 2 * sigma^2).
    """
    def __init__(self, sigma=1.0):
        self.sigma = sigma
    def kernel(self, t, x):
        return kernel.kernel_gaussian(t, x,
                   self.sigma)

class KernelExponential(Kernel):
    """Exponential kernel, exp(-||t_i - x_j|| / 2 * sigma^2).
    """
    def __init__(self, sigma=1.0):
        self.sigma = sigma
    def kernel(self, t, x):
        return kernel.kernel_exponential(t, x,
                   self.sigma)

class KernelSigmoid(Kernel):
    """Sigmoid kernel, tanh(gamma t_i' x_j + b).
    """
    def __init__(self, gamma=1.0, b=1.0):
        self.gamma = gamma
        self.b = b
    def kernel(self, t, x):
        return kernel.kernel_sigmoid(t, x,
            self.gamma, self.b)



## PLS

## This code is written by Davide Albanese, <davide.albanese@gmail.com>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['PLS']

import numpy as np


class PLS:
    """Multivariate primal Partial Least Squares (PLS) 
    algorithm as described in [Taylor04]_.

    .. [Taylor04] J Shawe-Taylor and N Cristianini. Kernel Methods for Pattern Analysis.
    """

    def __init__(self, iters):
        """Initialization.

        :Parameters:
           iters : int (>= 1)
              number of iterations. iters should be <= min(N-1, P)
        """

        self._iters = iters
        self._xmean = None
        self._beta0 = None
        self._beta = None

    def learn(self, x, y):
        """Compute the regression coefficients.

        Parameters:
           x : 2d array_like object
              training data (N, P)
           y : 1d array_like object (N [,M])
              target values
        """

        xarr = np.array(x, dtype=np.float, copy=True)
        yarr = np.array(y, dtype=np.float, copy=True)

        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")

        if yarr.ndim > 2:
            raise ValueError("y must be an 1d array_like object")
    
        if xarr.shape[0] != yarr.shape[0]:
            raise ValueError("x, y shape mismatch")

        if yarr.ndim == 1:
            yarr = yarr.reshape(-1, 1)
        
        self._xmean =  np.mean(xarr, axis=0)
        self._beta0 =  np.mean(yarr, axis=0)
        xarr = xarr - self._xmean
        yarr = yarr - self._beta0
        
        u = np.empty((xarr.shape[1], self._iters), dtype=np.float)
        c = np.empty((yarr.shape[1], self._iters), dtype=np.float)
        p = np.empty((xarr.shape[1], self._iters), dtype=np.float)
        for i in range(self._iters):
            YX = np.dot(yarr.T, xarr)
            u[:, i] = YX[0] / np.linalg.norm(YX[0])

            if yarr.shape[1] > 1:
                uold = u[:, i] + 1
                while np.linalg.norm(u[:, i] - uold) > 0.001:
                    uold = u[:, i]
                    tu = np.dot(np.dot(YX.T, YX), u[:, i])
                    u[:, i] = tu / np.linalg.norm(tu)
            
            t = np.dot(xarr, u[:, i].reshape(-1, 1))
            tt = np.dot(t.T, t)
            c[:, i] = np.ravel(np.dot(yarr.T, t) / tt)
            p[:, i] = np.ravel(np.dot(xarr.T, t) / tt)
            xarr = xarr - np.dot(t, p[:, i].reshape(1, -1))         
            
        self._beta = np.dot(u, np.linalg.solve(np.dot(p.T, u), c.T))
        
        if yarr.shape[1] == 1:
            self._beta = np.ravel(self._beta)

    def pred(self, t):
        """Compute the predicted response(s).

        :Parameters:
           t : 1d or 2d array_like object ([M,] P)
              test data

        :Returns:
           p : integer or 1d numpy darray
              predicted response(s)
        """

        if self._beta is None:
            raise ValueError("no model computed; run learn()")

        tarr = np.asarray(t, dtype=np.float)
        
        if tarr.ndim > 2 or tarr.ndim < 1:
            raise ValueError("t must be an 1d or a 2d array_like object")
        
        try:
            p = np.dot(tarr - self._xmean, self._beta) + self._beta0
        except ValueError:
            raise ValueError("t, beta: shape mismatch")

        return p
    
    def beta(self):
        """Returns the regression coefficients. 
        
        beta is a (P) vector in the univariate case
        and a (P, M) matrix in the multivariate case,
        where M is the number of target outputs.
        """

        return self._beta

    def beta0(self):
        """Returns offset(s). 
        
        beta is a float in the univariate case,
        and a (M) vector in the multivariate case,
        where M is the number of target outputs.
        """

        return self._beta0



## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from kernel_class import *


__all__ = ['LDAC', 'DLDA', 'KFDAC']


class LDAC:
    """Linear Discriminant Analysis Classifier.
    """
    
    def __init__(self):
        """Initialization.
        """

        self._labels = None
        self._w = None
        self._bias = None
      
    def learn(self, x, y):
        """Learning method.

        :Parameters:
           x : 2d array_like object
              training data (N, P)
           y : 1d array_like object integer
              target values (N)
        """
        
        xarr = np.asarray(x, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)
        
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")

        self._labels = np.unique(yarr)
        k = self._labels.shape[0]

        if k < 2:
            raise ValueError("number of classes must be >= 2")     
        
        p = np.empty(k, dtype=np.float)
        mu = np.empty((k, xarr.shape[1]), dtype=np.float)
        cov = np.zeros((xarr.shape[1], xarr.shape[1]), dtype=np.float)

        for i in range(k):
            wi = (yarr == self._labels[i])
            p[i] = np.sum(wi) / float(xarr.shape[0])
            mu[i] = np.mean(xarr[wi], axis=0)
            xi = xarr[wi] - mu[i]
            cov += np.dot(xi.T, xi)
        cov /= float(xarr.shape[0] - k)
        covinv = np.linalg.inv(cov)
        
        self._w = np.empty((k, xarr.shape[1]), dtype=np.float)
        self._bias = np.empty(k, dtype=np.float)

        for i in range(k):           
            self._w[i] = np.dot(covinv, mu[i])
            self._bias[i] = - 0.5 * np.dot(mu[i], self._w[i]) + \
                np.log(p[i])

    def labels(self):
        """Outputs the name of labels.
        """
        
        return self._labels
        
    def w(self):
        """Returns the coefficients.
        For multiclass classification this method returns a 2d 
        numpy array where w[i] contains the coefficients of label i.
        For binary classification an 1d numpy array (w_1 - w_0) 
        is returned.
        """
        
        if self._w is None:
            raise ValueError("no model computed.")

        if self._labels.shape[0] == 2:
            return self._w[1] - self._w[0]
        else:
            return self._w

    def bias(self):
        """Returns the bias.
        For multiclass classification this method returns a 1d 
        numpy array where b[i] contains the coefficients of label i. 
        For binary classification an float (b_1 - b_0) is returned.
        """
        
        if self._w is None:
            raise ValueError("no model computed.")
        
        if self._labels.shape[0] == 2:
            return self._bias[1] - self._bias[0]
        else:
            return self._bias

    def pred(self, t):
        """Does classification on test vector(s) `t`.
      
        :Parameters:
            t : 1d (one sample) or 2d array_like object
                test data ([M,] P)
            
        :Returns:        
            p : integer or 1d numpy array
                predicted class(es)
        """
        
        if self._w is None:
            raise ValueError("no model computed.")

        tarr = np.asarray(t, dtype=np.float)

        if tarr.ndim == 1:
            delta = np.empty(self._labels.shape[0], dtype=np.float)
            for i in range(self._labels.shape[0]):
                delta[i] = np.dot(tarr, self._w[i]) + self._bias[i]
            return self._labels[np.argmax(delta)]
        else:
            delta = np.empty((tarr.shape[0], self._labels.shape[0]),
                        dtype=np.float)
            for i in range(self._labels.shape[0]):
                delta[:, i] = np.dot(tarr, self._w[i]) + self._bias[i]
            return self._labels[np.argmax(delta, axis=1)]



class DLDA:
    """Diagonal Linear Discriminant Analysis classifier.
    The algorithm uses the procedure called Nearest Shrunken
    Centroids (NSC).
    """
    
    def __init__(self, delta):
        """Initialization.
        
        :Parameters:
           delta : float
              regularization parameter
        """

        self._delta = float(delta)
        self._xstd = None # s_j
        self._dprime = None # d'_kj
        self._xmprime = None # xbar'_kj
        self._p = None # class prior probability
        self._labels = None

    def learn(self, x, y):
        """Learning method.

        :Parameters:
           x : 2d array_like object
              training data (N, P)
           y : 1d array_like object integer
              target values (N)
        """
        
        xarr = np.asarray(x, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)
        
        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")
        
        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")

        self._labels = np.unique(yarr)
        k = self._labels.shape[0]

        if k < 2:
            raise ValueError("number of classes must be >= 2")
        
        xm = np.mean(xarr, axis=0)
        self._xstd = np.std(xarr, axis=0, ddof=1)
        s0 = np.median(self._xstd)
        self._dprime = np.empty((k, xarr.shape[1]), dtype=np.float)
        self._xmprime = np.empty((k, xarr.shape[1]), dtype=np.float)
        n = yarr.shape[0]
        self._p = np.empty(k, dtype=np.float)

        for i in range(k):
            yi = (yarr == self._labels[i])
            xim = np.mean(xarr[yi], axis=0)
            nk = np.sum(yi)
            mk = np.sqrt(nk**-1 - n**-1)
            d = (xim - xm) / (mk * (self._xstd + s0))
            
            # soft thresholding
            tmp = np.abs(d) - self._delta
            tmp[tmp<0] = 0.0
            self._dprime[i] = np.sign(d) * tmp
            
            self._xmprime[i] = xm + (mk * (self._xstd + s0) * self._dprime[i])
            self._p[i] = float(nk) / float(n)

    def labels(self):
        """Outputs the name of labels.
        """
        
        return self._labels
        
    def sel(self):
        """Returns the most important features (the features that 
        have a nonzero dprime for at least one of the classes).
        """

        return np.where(np.sum(self._dprime, axis=0) != 0)[0]

    def dprime(self):
        """Return the dprime d'_kj (C, P), where C is the
        number of classes.
        """
        
        return self._dprime

    def _score(self, x):
        """Return the discriminant score"""

        return - np.sum((x-self._xmprime)**2/self._xstd**2,
                        axis=1) + (2 * np.log(self._p))

    def _prob(self, x):
        """Return the probability estimates"""
        
        score = self._score(x)
        tmp = np.exp(score * 0.5)
        return tmp / np.sum(tmp)
        
    def pred(self, t):
        """Does classification on test vector(s) t.
      
        :Parameters:
           t : 1d (one sample) or 2d array_like object
              test data ([M,] P)
            
        :Returns:        
           p : int or 1d numpy array
              the predicted class(es) for t is returned.
        """
        
        if self._xmprime is None:
            raise ValueError("no model computed.")
        
        tarr = np.asarray(t, dtype=np.float)
        
        if tarr.ndim == 1:
            return self._labels[np.argmax(self._score(tarr))]
        else:
            ret = np.empty(tarr.shape[0], dtype=np.int)
            for i in range(tarr.shape[0]):
                ret[i] = self._labels[np.argmax(self._score(tarr[i]))]
            return ret
        
    def prob(self, t):
        """For each sample returns C (number of classes)
        probability estimates.
        """

        if self._xmprime is None:
            raise ValueError("no model computed.")
        
        tarr = np.asarray(t, dtype=np.float)

        if tarr.ndim == 1:
            return self._prob(tarr)
        else:
            ret = np.empty((tarr.shape[0], self._labels.shape[0]),
                dtype=np.float)
            for i in range(tarr.shape[0]):
                ret[i] = self._prob(tarr[i])
            return ret


class KFDAC:
    """Kernel Fisher Discriminant Analysis Classifier (binary classifier).

    The bias term (b) is computed as in [Gavin03]_.

    .. [Gavin03] Gavin C. et al. Efficient Cross-Validation of Kernel Fisher Discriminant Classifers. ESANN'2003 proceedings - European Symposium on Artificial Neural Networks, 2003.
    """
    
    def __init__(self, lmb=0.001, kernel=None):
        """Initialization.

        :Parameters:
           lmb : float (>= 0.0)
              regularization parameter
           kernel : None or mlpy.Kernel object.
              if kernel is None, K and Kt in .learn()
              and in .transform() methods must be precomputed kernel 
              matricies, else K and Kt must be training (resp. 
              test) data in input space.
        """

        if kernel is not None:
            if not isinstance(kernel, Kernel):
                raise ValueError("kernel must be None or a mlpy.Kernel object")

        self._lmb = float(lmb)
        self._kernel = kernel
        self._labels = None
        self._alpha = None
        self._b = None
        self._x = None
      
    def learn(self, K, y):
        """Learning method.

        :Parameters:
           K: 2d array_like object
              precomputed training kernel matrix (if kernel=None);
              training data in input space (if kernel is a Kernel object)
           y : 1d array_like object integer (N)
              class labels (only two classes)
        """

        Karr = np.asarray(K, dtype=np.float)
        yarr = np.asarray(y, dtype=np.int)

        if Karr.ndim != 2:
            raise ValueError("K must be a 2d array_like object")

        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")
        
        if Karr.shape[0] != yarr.shape[0]:
            raise ValueError("K, y shape mismatch")

        if self._kernel is None:
            if Karr.shape[0] != Karr.shape[1]:
                raise ValueError("K must be a square matrix")
        else:
            self._x = Karr.copy()
            Karr = self._kernel.kernel(Karr, Karr)

        self._labels = np.unique(yarr)
        if self._labels.shape[0] != 2:
            raise ValueError("number of classes must be = 2")
        
        n = yarr.shape[0]
        
        idx1 = np.where(yarr==self._labels[0])[0]
        idx2 = np.where(yarr==self._labels[1])[0]
        n1 = idx1.shape[0]
        n2 = idx2.shape[0]
        
        K1, K2 = Karr[:, idx1], Karr[:, idx2]
        
        N1 = np.dot(np.dot(K1, np.eye(n1) - (1 / float(n1))), K1.T)
        N2 = np.dot(np.dot(K2, np.eye(n2) - (1 / float(n2))), K2.T)
        N = N1 + N2 + np.diag(np.repeat(self._lmb, n))
        Ni = np.linalg.inv(N)

        m1 = np.sum(K1, axis=1) / float(n1)
        m2 = np.sum(K2, axis=1) / float(n2)
        d = (m1 - m2)
        M = np.dot(d.reshape(-1, 1), d.reshape(1, -1))

        self._alpha = np.linalg.solve(N, d)
        self._b = - np.dot(self._alpha, (n1 * m1 + n2 * m2) / float(n))

    def labels(self):
        """Outputs the name of labels.
        """
        
        return self._labels
        
    def pred(self, Kt):
        """Compute the predicted response.
      
        :Parameters:
           Kt : 1d or 2d array_like object
              precomputed test kernel matrix. (if kernel=None);
              test data in input space (if kernel is a Kernel object).
            
        :Returns:        
            p : integer or 1d numpy array
                the predicted class(es)
        """

        
        if self._alpha is None:
            raise ValueError("no model computed; run learn()")

        Ktarr = np.asarray(Kt, dtype=np.float)
        if self._kernel is not None:
            Ktarr = self._kernel.kernel(Ktarr, self._x)

        try:
            s = np.sign(np.dot(self._alpha, Ktarr.T) + self._b)
        except ValueError:
            raise ValueError("Kt, alpha: shape mismatch")

        return np.where(s==1, self._labels[0], self._labels[1]) \
            .astype(np.int)

    def alpha(self):
        """Return alpha.
        """
        return self._alpha

    def b(self):
        """Return b.
        """
        return self._b



## This code is written by Davide Albanese, <davide.albanese@gmail.com>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['pad']

import numpy as np


def next_p2(n):
    """Returns the smallest integer, greater than n
    (n positive and >= 1) which can be obtained as power of 2.
    """
    
    if n < 1:
        raise ValueError("n must be >= 1")

    v = 2
    while v <= n:
        v = v * 2

    return v


def pad(x, method='reflection'):
    """Pad to bring the total length N up to the next-higher 
    power of two.

    :Parameters:
       x : 1d array_like object
          data
       method : string ('reflection', 'periodic', 'zeros')
          method
          
    :Returns:
       xp, orig : 1d numpy array, 1d numpy array bool
          padded version of `x` and a boolean array with
          value True where xp contains the original data
    """

    x_arr = np.asarray(x)

    if not method in ['reflection', 'periodic', 'zeros']:
        raise ValueError('method %s not available' % method)
    
    diff = next_p2(x_arr.shape[0]) - x_arr.shape[0]
    ldiff = int(diff / 2)
    rdiff = diff - ldiff

    if method == 'reflection':
        left_x = x_arr[:ldiff][::-1]
        right_x = x_arr[-rdiff:][::-1]         
    elif method == 'periodic':
        left_x = x_arr[:ldiff]
        right_x = x_arr[-rdiff:]
    elif method == 'zeros':
        left_x = np.zeros(ldiff, dtype=x_arr.dtype)
        right_x = np.zeros(rdiff, dtype=x_arr.dtype)
        
    xp = np.concatenate((left_x, x_arr, right_x))
    orig = np.ones(x_arr.shape[0] + diff, dtype=np.bool)
    orig[:ldiff] = False
    orig[-rdiff:] = False
   
    return xp, orig



"""Wavelet transform
"""

import sys
if sys.version >= '3':
    from ._dwt import *
    from ._uwt import *
else:
    from _dwt import *
    from _uwt import *

from continuous import *
import continuous
from uwt_align import *
import uwt_align
from padding import *
import padding


__all__ = []
__all__ += continuous.__all__
__all__ += uwt_align.__all__
__all__ += ["dwt", "idwt", "uwt", "iuwt"]
__all__ += padding.__all__



from distutils.core import setup, Extension
from distutils.sysconfig import *
from distutils.util import *
import os
import os.path
import numpy

try:
   from distutils.command.build_py import build_py_2to3 \
       as build_py
except ImportError:
   from distutils.command.build_py import build_py
   
try:
   from Cython.Distutils import build_ext
except ImportError:
   use_cython = False
else:
   use_cython = True
   
   
#### data files
data_files = []
# Include gsl libs for the win32 distribution
if get_platform() == "win32":
   dlls = ["gsl.lib", "cblas.lib"]
   data_files += [("Lib\site-packages\mlpy", dlls)]
   
#### libs
if get_platform() == "win32":
   gsl_lib = ['gsl', 'cblas']
   math_lib = []
else:
   gsl_lib = ['gsl', 'gslcblas']
   math_lib = ['m']
   
#### Extra compile args
if get_platform() == "win32":
   extra_compile_args = []
else:
   extra_compile_args = ['-Wno-strict-prototypes']
   
#### Python include
py_inc = [get_python_inc()]

#### NumPy include
np_lib = os.path.dirname(numpy.__file__)
np_inc = [os.path.join(np_lib, 'core/include')]

#### scripts
scripts = []

#### cmdclass
cmdclass = {'build_py': build_py}

#### Extension modules
ext_modules = []
if use_cython:
   cmdclass.update({'build_ext': build_ext})
   ext_modules += [Extension("mlpy.gsl", ["mlpy/gsl/gsl.pyx"],
                             libraries=gsl_lib + math_lib,
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.liblinear",
                             ["mlpy/liblinear/liblinear/linear.cpp",
                              "mlpy/liblinear/liblinear/tron.cpp",
                              "mlpy/liblinear/liblinear.pyx",
                              "mlpy/liblinear/liblinear/blas/daxpy.c",
                              "mlpy/liblinear/liblinear/blas/ddot.c",
                              "mlpy/liblinear/liblinear/blas/dnrm2.c",
                              "mlpy/liblinear/liblinear/blas/dscal.c"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.libsvm",
                             ["mlpy/libsvm/libsvm/svm.cpp",
                              "mlpy/libsvm/libsvm.pyx"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.libml",
                             ["mlpy/libml/src/alloc.c",
                              "mlpy/libml/src/dist.c",
                              "mlpy/libml/src/get_line.c",
                              "mlpy/libml/src/matrix.c",
                              "mlpy/libml/src/mlg.c",
                              "mlpy/libml/src/nn.c",
                              "mlpy/libml/src/parser.c",
                              "mlpy/libml/src/read_data.c",
                              "mlpy/libml/src/rn.c",
                              "mlpy/libml/src/rsfn.c",
                              "mlpy/libml/src/sampling.c",
                              "mlpy/libml/src/sort.c",
                              "mlpy/libml/src/svm.c",
                              "mlpy/libml/src/tree.c",
                              "mlpy/libml/src/trrn.c",
                              "mlpy/libml/src/ttest.c",
                              "mlpy/libml/src/unique.c",
                              "mlpy/libml/libml.pyx"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.kmeans",
                             ["mlpy/kmeans/c_kmeans.c",
                              "mlpy/kmeans/kmeans.pyx"],
                             include_dirs=py_inc + np_inc,
                             libraries=gsl_lib + math_lib),
                   Extension("mlpy.kernel",
                             ["mlpy/kernel/c_kernel.c",
                              "mlpy/kernel/kernel.pyx"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.canberra",
                             ["mlpy/canberra/c_canberra.c",
                              "mlpy/canberra/canberra.pyx"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.adatron",
                             ["mlpy/adatron/c_adatron.c",
                              "mlpy/adatron/adatron.pyx"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.findpeaks",
                             ["mlpy/findpeaks/c_findpeaks.c",
                              "mlpy/findpeaks/findpeaks.pyx"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.dtw", 
                             ["mlpy/dtw/cdtw.c",
                             "mlpy/dtw/dtw.pyx"],
                             libraries=math_lib,
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.lcs", 
                             ["mlpy/lcs/clcs.c",
                              "mlpy/lcs/lcs.pyx"],
                             libraries=math_lib,
                             include_dirs=py_inc + np_inc),
                   ]
else:
   ext_modules += [Extension("mlpy.gsl", ["mlpy/gsl/gsl.c"],
                             libraries=gsl_lib + math_lib,
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.liblinear",
                             ["mlpy/liblinear/liblinear/linear.cpp",
                              "mlpy/liblinear/liblinear/tron.cpp",
                              "mlpy/liblinear/liblinear.c",
                              "mlpy/liblinear/liblinear/blas/daxpy.c",
                              "mlpy/liblinear/liblinear/blas/ddot.c",
                              "mlpy/liblinear/liblinear/blas/dnrm2.c",
                              "mlpy/liblinear/liblinear/blas/dscal.c"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.libsvm",
                             ["mlpy/libsvm/libsvm/svm.cpp",
                              "mlpy/libsvm/libsvm.c"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.libml",
                             ["mlpy/libml/src/alloc.c",
                              "mlpy/libml/src/dist.c",
                              "mlpy/libml/src/get_line.c",
                              "mlpy/libml/src/matrix.c",
                              "mlpy/libml/src/mlg.c",
                              "mlpy/libml/src/nn.c",
                              "mlpy/libml/src/parser.c",
                              "mlpy/libml/src/read_data.c",
                              "mlpy/libml/src/rn.c",
                              "mlpy/libml/src/rsfn.c",
                              "mlpy/libml/src/sampling.c",
                              "mlpy/libml/src/sort.c",
                              "mlpy/libml/src/svm.c",
                              "mlpy/libml/src/tree.c",
                              "mlpy/libml/src/trrn.c",
                              "mlpy/libml/src/ttest.c",
                              "mlpy/libml/src/unique.c",
                              "mlpy/libml/libml.c"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.kmeans",
                             ["mlpy/kmeans/c_kmeans.c",
                              "mlpy/kmeans/kmeans.c"],
                             include_dirs=py_inc + np_inc,
                             libraries=gsl_lib + math_lib),
                   Extension("mlpy.kernel",
                             ["mlpy/kernel/c_kernel.c",
                              "mlpy/kernel/kernel.c"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.canberra",
                             ["mlpy/canberra/c_canberra.c",
                              "mlpy/canberra/canberra.c"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.adatron",
                             ["mlpy/adatron/c_adatron.c",
                              "mlpy/adatron/adatron.c"],
                             include_dirs=py_inc + np_inc,
                             libraries=math_lib),
                   Extension("mlpy.findpeaks",
                             ["mlpy/findpeaks/c_findpeaks.c",
                              "mlpy/findpeaks/findpeaks.c"],
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.dtw",
                             ["mlpy/dtw/cdtw.c",
                              "mlpy/dtw/dtw.c"],
                             libraries=math_lib,
                             include_dirs=py_inc + np_inc),
                   Extension("mlpy.lcs",
                             ["mlpy/lcs/clcs.c",
                              "mlpy/lcs/lcs.c"],
                             libraries=math_lib,
                             include_dirs=py_inc + np_inc),
                   ]
   
# pure c modules
ext_modules += [Extension("mlpy.wavelet._dwt",
                          ["mlpy/wavelet/dwt.c"],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc,
                          libraries=gsl_lib + math_lib),
                Extension("mlpy.wavelet._uwt",
                          ["mlpy/wavelet/uwt.c"],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc,
                          libraries=gsl_lib + math_lib),
                Extension("mlpy.hcluster.chc",
                          ["mlpy/hcluster/hc.c"],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc,
                          libraries=math_lib),
                Extension("mlpy.bordacount.cborda",
                          ["mlpy/bordacount/borda.c"],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc),
                Extension("mlpy.bordacount.cborda",
                          ["mlpy/bordacount/borda.c"],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc),
                Extension('mlpy.fastcluster._fastcluster',
                          ['mlpy/fastcluster/fastcluster/src/fastcluster_python.cpp'],
                          extra_compile_args=extra_compile_args,
                          include_dirs=py_inc + np_inc),
                ]

packages=['mlpy', 'mlpy.wavelet', 'mlpy.hcluster',
          'mlpy.bordacount', 'mlpy.fastcluster']

classifiers = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: GNU General Public License (GPL)',
               'Programming Language :: C',
               'Programming Language :: C++',
               'Programming Language :: Python',
               'Programming Language :: Python :: 3',
               'Topic :: Scientific/Engineering :: Artificial Intelligence',
               'Topic :: Scientific/Engineering :: Mathematics',
               'Operating System :: POSIX :: Linux',
               'Operating System :: POSIX :: BSD',
               'Operating System :: MacOS',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX'
               ]

setup(name = 'mlpy',
      version='3.5.0',
      requires=['numpy (>=1.3.0)', 'scipy (>=0.7.0)', 'gsl (>=1.11)'],
      description='mlpy - Machine Learning Py - ' \
         'High-Performance Python Package for Predictive Modeling',
      author='mlpy Developers',
      author_email='davide.albanese@gmail.com',
      maintainer='Davide Albanese',
      maintainer_email='davide.albanese@gmail.com',
      packages=packages,
      url='mlpy.sourceforge.net',
      download_url='https://sourceforge.net/projects/mlpy/',
      license='GPLv3',
      classifiers=classifiers,
      cmdclass=cmdclass,
      ext_modules=ext_modules,
      scripts=scripts,
      data_files=data_files
      )



## This code is written by Davide Albanese, <albanese@fbk.eu>
## (C) 2010 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
cimport numpy as np
from libc.stdlib cimport *

from cliblinear cimport *

    
cdef void print_null(char *s):
    pass

# array 1D to feature node
cdef feature_node *array1d_to_node(np.ndarray[np.float64_t, ndim=1] x):
    cdef int i, k
    cdef np.ndarray[np.int32_t, ndim=1] nz
    cdef feature_node *ret

    nz = np.nonzero(x)[0].astype(np.int32)
    nf = nz.shape[0] + 2 # bias term + last (-1) term
        
    ret = <feature_node*> malloc (nf * sizeof(feature_node))

    k = 0
    for i in nz:
        ret[k].index = i+1
        ret[k].value = x[i]
        k += 1

    ret[nz.shape[0]].index = x.shape[0] + 1 # add bias term
    ret[nz.shape[0]].value = 1 # add bias term
    ret[nz.shape[0]+1].index = -1 # last term
                
    return ret

# array 2D to svm node
cdef feature_node **array2d_to_node(np.ndarray[np.float64_t, ndim=2] x):
    cdef int i
    cdef feature_node **ret

    ret = <feature_node **> malloc \
        (x.shape[0] * sizeof(feature_node *))
    
    for i in range(x.shape[0]):
        ret[i] = array1d_to_node(x[i])
            
    return ret

# array 1D to vector
cdef int *array1d_to_vector(np.ndarray[np.int32_t, ndim=1] y):
     cdef int i
     cdef int *ret

     ret = <int *> malloc (y.shape[0] * sizeof(int))

     for i in range(y.shape[0]):
         ret[i] = y[i]

     return ret


cdef class LibLinear:
    cdef problem problem
    cdef parameter parameter
    cdef model *model
    cdef int learn_disabled
        
    SOLVER_TYPE = ['l2r_lr',
                   'l2r_l2loss_svc_dual',
                   'l2r_l2loss_svc',
                   'l2r_l1loss_svc_dual',
                   'mcsvm_cs',
                   'l1r_l2loss_svc',
                   'l1r_lr',
                   'l2r_lr_dual']

    def __cinit__(self, solver_type='l2r_lr', C=1, eps=0.01, weight={}):
        """LibLinear is a simple class for solving large-scale regularized
        linear classification. It currently supports L2-regularized logistic
        regression/L2-loss support vector classification/L1-loss support vector
        classification, and L1-regularized L2-loss support vector classification/
        logistic regression.
                
        :Parameters:
            solver_type : string
                solver, can be one of 'l2r_lr', 'l2r_l2loss_svc_dual',
                'l2r_l2loss_svc', 'l2r_l1loss_svc_dual', 'mcsvm_cs',
                'l1r_l2loss_svc', 'l1r_lr', 'l2r_lr_dual'
            C : float
                cost of constraints violation
            eps : float
                stopping criterion
            weight : dict 
                changes the penalty for some classes (if the weight for a
                class is not changed, it is set to 1). For example, to
                change penalty for classes 1 and 2 to 0.5 and 0.8
                respectively set weight={1:0.5, 2:0.8}
        """

        set_print_string_function(&print_null)
        self.learn_disabled = 0

        try:
            self.parameter.solver_type = self.SOLVER_TYPE.index(solver_type)
        except ValueError:
            raise ValueError("invalid solver_type")

        self.parameter.C = C
        self.parameter.eps = eps
                
        # weight
        self.parameter.nr_weight = len(weight)
        self.parameter.weight_label = <int *> malloc \
            (len(weight) * sizeof(int))
        self.parameter.weight = <double *> malloc \
            (len(weight) * sizeof(double))
        try:
            for i, key in enumerate(weight):
                self.parameter.weight_label[i] = int(key)
                self.parameter.weight[i] = float(weight[key])
        except ValueError:
            raise ValueError("invalid weight")

        self.model = NULL

    def __dealloc__(self):
        self._free_problem()
        self._free_model()
        self._free_param()
        
    def _load_problem(self, x, y):
        """Convert the data into liblinear problem struct
        """
        
        xarr = np.ascontiguousarray(x, dtype=np.float64)
        yarr = np.ascontiguousarray(y, dtype=np.int32)

        if xarr.ndim != 2:
            raise ValueError("x must be a 2d array_like object")

        if yarr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")

        if xarr.shape[0] != yarr.shape[0]:
            raise ValueError("x, y: shape mismatch")
        
        self.problem.x = array2d_to_node(xarr)
        self.problem.y = array1d_to_vector(yarr)
        self.problem.l = xarr.shape[0]
        self.problem.n = xarr.shape[1] + 1 # + 1 for the bias term
        self.problem.bias = 1
        
    def learn(self, x, y):
        """Learning method.
        
        :Parameters:
      
            x : 2d array_like object
                training data (N, P)
            y : 1d array_like object
                target values (N)
        """

        cdef char *ret

        srand(1)

        if self.learn_disabled:
            raise ValueError("learn method is disabled (model from file)")
        
        self._free_problem()
        self._load_problem(x, y)
        ret = check_parameter(&self.problem, &self.parameter)

        if ret != NULL:
            raise ValueError(ret)

        self._free_model()
        self.model = train(&self.problem, &self.parameter)

    def pred(self, t):
        """Does classification on test vector(s) t.
                
        :Parameters:
            t : 1d (one sample) or 2d array_like object
                test data ([M,] P)
            
        :Returns:
            p : int or 1d numpy array
                the predicted class(es) for t is returned.
        """

        cdef int i
        cdef feature_node *test_node

        tarr = np.ascontiguousarray(t, dtype=np.float64)

        if tarr.ndim > 2:
            raise ValueError("t must be an 1d or a 2d array_like object")
        
        if self.model is NULL:
            raise ValueError("no model computed")

        if tarr.ndim == 1:
            test_node = array1d_to_node(tarr)
            p = predict(self.model, test_node)
            free(test_node)
        else:
            p = np.empty(tarr.shape[0], dtype=np.int)
            for i in range(tarr.shape[0]):
                test_node = array1d_to_node(tarr[i])
                p[i] = predict(self.model, test_node)
                free(test_node)
        
        return p


    def pred_values(self, t):
        """Returns D decision values. D is 1 if there are two 
        classes except multi-class svm by Crammer and Singer 
        ('mcsvm_cs'), and is the number of classes otherwise.
        The pred() method returns the class with the highest 
        decision value.
                
        :Parameters:
            t : 1d (one sample) or 2d array_like object
                test data ([M,] P)
            
        :Returns:
            decision values : 1d (D) or 2d numpy array (M, D)
                decision values for each observation.
        """

        cdef int i, j
        cdef feature_node *test_node
        cdef double *dec_values

        tarr = np.ascontiguousarray(t, dtype=np.float64)

        if tarr.ndim > 2:
            raise ValueError("t must be an 1d or a 2d array_like object")
        
        if self.model is NULL:
            raise ValueError("no model computed")

        if (self.SOLVER_TYPE[self.parameter.solver_type] != \
                'mcsvm_cs') and self.model.nr_class == 2:
            n = 1
        else:
            n = self.model.nr_class

        dec_values = <double *> malloc (n * sizeof(double))

        if tarr.ndim == 1:
            dec_values_arr = np.empty(n, dtype=np.float)
            test_node = array1d_to_node(tarr)
            p = predict_values(self.model, test_node, dec_values)
            free(test_node)
            for j in range(n):
                dec_values_arr[j] = dec_values[j]
        else:
            dec_values_arr = np.empty((tarr.shape[0], n), dtype=np.float)
            for i in range(tarr.shape[0]):
                test_node = array1d_to_node(tarr[i])
                p = predict_values(self.model, test_node, dec_values)
                free(test_node)
                for j in range(n):
                    dec_values_arr[i, j] = dec_values[j]
        
        free(dec_values)
        return dec_values_arr


    def pred_probability(self, t):
        """Returns C (number of classes) probability estimates. 
        The simple probability model of logistic regression
        is used.

        :Parameters:
            t : 1d (one sample) or 2d array_like object
                test data ([M,] P)
            
        :Returns:
            probability estimates : 1d (C) or 2d numpy array (M, C)
                probability estimates for each observation.
        """

        cdef int i, j
        cdef feature_node *test_node
        cdef double *prob_estimates

        tarr = np.ascontiguousarray(t, dtype=np.float64)

        if tarr.ndim > 2:
            raise ValueError("t must be an 1d or a 2d array_like object")
        
        if self.model is NULL:
            raise ValueError("no model computed")

        prob_estimates = <double *> malloc (self.model.nr_class *
                                            sizeof(double))

        if tarr.ndim == 1:
            prob_estimates_arr = np.empty(self.model.nr_class, dtype=np.float)
            test_node = array1d_to_node(tarr)
            p = predict_probability(self.model, test_node, prob_estimates)
            free(test_node)
            for j in range(self.model.nr_class):
                prob_estimates_arr[j] = prob_estimates[j]
        else:
            prob_estimates_arr = np.empty((tarr.shape[0], self.model.nr_class), 
                                      dtype=np.float)
            for i in range(tarr.shape[0]):
                test_node = array1d_to_node(tarr[i])
                p = predict_probability(self.model, test_node, prob_estimates)
                free(test_node)
                for j in range(self.model.nr_class):
                    prob_estimates_arr[i, j] = prob_estimates[j]
        
        free(prob_estimates)
        return prob_estimates_arr


    def nfeature(self):
        """Returns the number of attributes.
        """

        if self.model is NULL:
            raise ValueError("no model")

        return self.model.nr_feature

    def nclasses(self):
        """Returns the number of classes.
        """
        
        if self.model is NULL:
            raise ValueError("no model")

        return self.model.nr_class

    def labels(self):
        """Outputs the name of labels.
        """
        
        if self.model is NULL:
            raise ValueError("no model computed")

        if self.model.label is NULL:
            return None
        else:
            ret = np.empty(self.model.nr_class, dtype=np.int32)
            for i in range(self.model.nr_class):
                ret[i] = self.model.label[i]
            return ret
    
    def _w(self):
               
        cdef int i, n
        cdef np.ndarray[np.float64_t, ndim=1] w

        if self.model is NULL:
            raise ValueError("no model")

        if (self.SOLVER_TYPE[self.parameter.solver_type] == \
                'mcsvm_cs') or self.model.nr_class != 2:
            n = self.problem.n * self.model.nr_class
        else:
            n = self.problem.n
            
        w = np.empty(n, dtype=np.float64)

        for i in range(n):
            w[i] = self.model.w[i]

        if (self.SOLVER_TYPE[self.parameter.solver_type] == \
                'mcsvm_cs') or self.model.nr_class != 2:
            return w.reshape(-1, self.model.nr_class).T
        else:
            return w

    def w(self):
        """Returns the coefficients.
        For 'mcsvm_cs' solver and for multiclass classification 
        returns a 2d numpy array where w[i] contains the
        coefficients of label i. For binary classification 
        an 1d numpy array is returned.
        """

        w = self._w()
        
        if w.ndim == 1:
            return w[:-1]
        else:
            return w[:, :-1]
       
    def bias(self):
        """Returns the bias term(s).
        For 'mcsvm_cs' solver and for multiclass classification
        returns a 1d numpy array where b[i] contains the
        bias of label i (.labels()[i]). For binary classification 
        a float is returned.
        """

        w = self._w()
        
        if w.ndim == 1:
            return w[-1]
        else:
            return w[:, -1]
        
    cdef void _free_problem(self):
        cdef int i
        
        if self.problem.x is not NULL:
            for i in range(self.problem.l):
                free(self.problem.x[i])
            free(self.problem.x)

        if self.problem.y is not NULL:
            free(self.problem.y)
        
    cdef void _free_model(self):
        free_and_destroy_model(&self.model)

    cdef void _free_param(self):
        destroy_param(&self.parameter)
    
    @classmethod
    def load_model(cls, filename):
        """Loads model from file. Returns a LibLinear object
        with the learn() method disabled.
        """
        
        ret = LibLinear()

        try:
            ret.model = load_model(filename)
        except:
            raise ValueError("invalid filename")
        else:
            if ret.model is NULL:
                raise IOError("model file could not be loaded")
            
        ret.parameter.solver_type = ret.model.param.solver_type
        ret.learn_disabled = 1
        
        return ret

    def save_model(self, filename):
        """Saves a model to a file.        
        """
        
        cdef int ret
        
        if self.model is NULL:
            raise ValueError("no model computed")
        
        ret = save_model(filename, self.model)
        
        if ret == -1:
            raise IOError("problem with save_model()")



## This code is written by Davide Albanese, <albanese@fbk.eu>
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
cimport numpy as np

cdef extern from "c_adatron.h":
    int adatron(long *y, double *K, int n, double C, int maxsteps, 
                double eps,double *alpha, double *margin)
    
np.import_array()
    

class KernelAdatron:
    """Kernel Adatron algorithm without-bias-term (binary classifier).
    
    The algoritm handles a version of the 1-norm soft margin
    support vector machine. If C is very high the algoritm 
    handles a version of the hard margin SVM.

    Use positive definite kernels (such as Gaussian
    and Polynomial kernels)
    """
    
    def __init__(self, C=1000, maxsteps=1000, eps=0.01):
        """Initialization.

        :Parameters:
           C : float
              upper bound on the value of alpha
           maxsteps : integer (> 0)
              maximum number of steps
           eps : float (>=0)
              the algoritm stops when abs(1 - margin) < eps
        """

        self._C = float(C)
        self._maxsteps = int(maxsteps)
        self._eps = float(eps)
        self._alpha = None
        self._margin = None
        self._steps = None
        self._labels = None
        self._y = None

    def learn(self, K, y):
        """Learn.

        Parameters:
           K: 2d array_like object (N, N)
              precomputed kernel matrix
           y : 1d array_like object (N)
              target values
        """
        
        cdef np.ndarray[np.int_t, ndim=1] ynew
        cdef np.ndarray[np.float_t, ndim=2] K_arr
        cdef np.ndarray[np.float_t, ndim=1] alpha_arr
        cdef double margin

        K_arr = np.ascontiguousarray(K, dtype=np.float)
        y_arr = np.asarray(y, dtype=np.int)

        if K_arr.ndim != 2:
            raise ValueError("K must be a 2d array_like object")

        if y_arr.ndim != 1:
            raise ValueError("y must be an 1d array_like object")
        
        if K_arr.shape[0] != K_arr.shape[1]:
            raise ValueError("K must be a square matrix")

        if K_arr.shape[0] != y_arr.shape[0]:
            raise ValueError("K, y shape mismatch")
        
        self._labels = np.unique(y_arr)
        if self._labels.shape[0] != 2:
            raise ValueError("number of classes != 2")

        ynew = np.where(y_arr==self._labels[0], -1, 1)
        n = K_arr.shape[0]
        alpha_arr = np.zeros(n, dtype=np.float)

        steps = adatron(<long*> ynew.data, <double *> K_arr.data,
            <int> n, <double> self._C, <int> self._maxsteps,
            <double> self._eps, <double *> alpha_arr.data, &margin)

        self._alpha = alpha_arr
        self._margin = margin
        self._steps = steps
        self._y = ynew

    def pred(self, Kt):
        """Compute the predicted class.

        :Parameters:
           Kt : 1d or 2d array_like object ([M], N)
              test kernel matrix. Precomputed inner products 
              (in feature space) between M testing and N 
              training points.

        :Returns:
           p : integer or 1d numpy array
              predicted class
        """

        if self._alpha is None:
            raise ValueError("no model computed; run learn() first")

        Kt_arr = np.asarray(Kt, dtype=np.float)

        try:
            s = np.sign(np.dot(self._alpha * self._y, Kt_arr.T))
        except ValueError:
            raise ValueError("Kt, alpha: shape mismatch")
        
        return np.where(s==-1, self._labels[0], self._labels[1]) \
            .astype(np.int)

    def margin(self):
        """Return the margin.
        """

        return self._margin
    
    def steps(self):
        """Return the number of steps performed.
        """

        return self._steps
    
    def alpha(self):
        """Return alpha
        """
        
        return self._alpha



from version import v as __version__
import sys

# extension modules
if sys.version >= '3':
    from . import gsl
    from .libsvm import *
    from .liblinear import *
    from .libml import *
    from .findpeaks import *
    from .kmeans import *
    from .kernel import *
    from .adatron import *
    from .canberra import *
    from .dtw import *
    from .lcs import *
else:
    import gsl
    from libsvm import *
    from liblinear import *
    from libml import *
    from findpeaks import *
    from kmeans import *
    from kernel import *
    from adatron import *
    from canberra import *
    from dtw import *
    from lcs import *

# python modules
from crossval import *
from hcluster import *
from metrics import *
from perceptron import *
from da import *
from ols import *
from ridge import *
from bordacount import *
from lars import *
from elasticnet import *
from dimred import *
from irelief import *
from parzen import *
from stats import *
from fastcluster import *
from kernel_class import *
from rfe import *
from golub import *
from pls import *

import crossval
import hcluster
import metrics
import perceptron
import da
import ols
import ridge
import bordacount
import lars
import elasticnet
import dimred
import irelief
import parzen
import stats
import fastcluster
import kernel_class
import rfe
import golub
import pls

# visible submodules
import wavelet


__all__ = []
__all__ += crossval.__all__
__all__ += hcluster.__all__
__all__ += metrics.__all__
__all__ += perceptron.__all__
__all__ += da.__all__
__all__ += ols.__all__
__all__ += ridge.__all__
__all__ += bordacount.__all__
__all__ += lars.__all__
__all__ += elasticnet.__all__
__all__ += dimred.__all__
__all__ += irelief.__all__
__all__ += parzen.__all__
__all__ += stats.__all__
__all__ += fastcluster.__all__
__all__ += kernel_class.__all__
__all__ += rfe.__all__
__all__ += golub.__all__
__all__ += pls.__all__

__all__ += ['LibLinear']
__all__ += ['LibSvm']
__all__ += ['findpeaks_dist', 'findpeaks_win']
__all__ += ['kmeans']
__all__ += ['kernel_linear', 'kernel_gaussian', 
            'kernel_polynomial', 'kernel_exponential', 
            'kernel_sigmoid', 'kernel_center']
__all__ += ['KernelAdatron']
__all__ += ['canberra', 'canberra_location', 'canberra_stability',
            'canberra_location_expected', 'canberra_location_max', 
            'canberra_stability_max']
__all__ += ['KNN', 'ClassTree', 'MaximumLikelihoodC']
__all__ += ['dtw_std', 'dtw_subsequence']
__all__ += ['lcs_std', 'lcs_real']



## Findpeaks submodule

## This code is written by Davide Albanese, <davide.albanese@gmail.com>.
## (C) 2011 mlpy Developers.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["findpeaks_dist", "findpeaks_win"]

import numpy as np
cimport numpy as np
cimport cython
from stdlib cimport free

cdef extern from "c_findpeaks.h":
   int *fp_win(double *x, int n, int span, int *m)


@cython.boundscheck(False)
@cython.wraparound(False)
def findpeaks_win(x, span):
    """Find peaks with a sliding window of
    width `span`.

    :Parameters:
        x : 1d array_like object
            input data
        span : odd integer (>=3)
            span
    
    :Returns:
        idx : 1d numpy array int
            peaks indexes

    Example:
    
    >>> import mlpy
    >>> x = [6,2,2,1,3,4,1,3,1,1,1,6,2,2,7,1]
    >>> mlpy.findpeaks_win(x, span=3)
    array([ 0,  5,  7, 11, 14])
    """

    cdef np.ndarray[np.float_t, ndim=1] xarr
    cdef np.ndarray[np.int_t, ndim=1] ret
    cdef int *retc
    cdef int i, m
    
    span = int(span)

    if (span % 2 == 0) or (span < 3):
        raise ValueError("span must be >= 3 and odd")
    
    xarr = np.ascontiguousarray(x, dtype=np.float)
    retc = fp_win(<double *> xarr.data, <int> xarr.shape[0], <int> span, &m)
    ret = np.empty(m, dtype=np.int)
    
    for i in range(m):
        ret[i] = retc[i]
            
    free(retc)
    return ret


@cython.boundscheck(False)
@cython.wraparound(False)
def findpeaks_dist(x, mindist=2):
    """Find peaks. With `mindist` parameter 
    the algorithm ignore small peaks that occur
    in the neighborhood of a larger peak.

    :Parameters:
        x : 1d array_like object
            input data
        mindist : integer (>=2)
            minimum peak distance (minimum separation
            between peaks)
    
    :Returns:
        idx : 1d numpy array int
            peaks indexes

    Example:
    
    >>> import mlpy
    >>> x = [6,2,2,1,3,4,1,3,1,1,1,6,2,2,7,1]
    >>> mlpy.findpeaks_dist(x, mindist=3)
    array([ 0,  5, 11, 14])
    """

    cdef long i, j, mi
    cdef double mm
    cdef np.ndarray[np.float_t, ndim=1] _x
    cdef np.ndarray[np.int_t, ndim=1] idx
    cdef np.ndarray[np.int_t, ndim=1] tmp
    
    if mindist < 2:
        raise ValueError("mindist must be >= 2")

    _x = np.ascontiguousarray(x, dtype=np.float)
    idx = findpeaks_win(_x, 3)
    tmp = np.empty_like(idx)
 
    j = 0
    mm = _x[idx[0]]
    mi = idx[0]
    
    for i in range(1, idx.shape[0]):
        if (idx[i] - idx[i-1]) < mindist:
            if _x[idx[i]] > mm:
                mm = _x[idx[i]]
                mi = idx[i]
        else:
            tmp[j] = mi
            j += 1
            mm = _x[idx[i]]
            mi = idx[i]
    
    tmp[j] = mi
    return tmp[:j+1]



cdef extern from "gsl/gsl_sf.h":

    double gsl_sf_gamma (double x)
    double gsl_sf_fact (unsigned int n)

cdef extern from "gsl/gsl_statistics_double.h":
    double gsl_stats_quantile_from_sorted_data(double sorted_data[], size_t stride, size_t n, double f)



from cgsl cimport *
import numpy as np
cimport numpy as np

def sf_gamma (double x):
    return gsl_sf_gamma (x)

def sf_fact (unsigned int n):
    return gsl_sf_fact (n)

def stats_quantile_from_sorted_data (sorted_data, f):
    cdef np.ndarray[np.float_t, ndim=1] sorted_data_arr

    try:
        sorted_data_arr = np.array(sorted_data, dtype=np.float, order='C')
    except ValueError:
        raise ValueError("sorted_data must be a 2d array_like object")

    return gsl_stats_quantile_from_sorted_data (<double *> 
        sorted_data_arr.data, 1, <size_t> sorted_data_arr.shape[0], 
        <double> f)
